package com.perisic.jshadow;
import java.lang.reflect.Method;
import java.util.Vector; 

/**
* Providing some utility methods, especially methods that add
* and remove 
* shadows to a class.
*<UL>
*<li> First version: 16.09.2004
*<li>
* Copyright: (c) Marc Conrad, 2004
*<li>
* See http://perisic.com/shadow for information about this package. 
*<li>The com.perisic.shadow library is distributed under the terms of the 
* <a href="http://www.gnu.org/copyleft/lesser.html" target="_blank">
GNU Lesser General Public License (LGPL)</a>. 
*<li> If you require the package under a different licence please contact me.
*<li> disclaimer: The classes are provided "as is".
* There is no warranty implied by using the com.perisic.shadow package.
* <li> Note: This class needs a special adapted version of the <a name="jikes">jikes</a> compiler (available on
* <a href="http://perisic.com/shadow">http://perisic.com/shadow</a>) 
* for succesfull compilation. Compile this 
*      class with the command jikes -dy
* </UL>
*/

public final class Util {

/**
* Generate no instance of that class. 
*/
private Util() { }
// Algorithm for finding the most specific method is 
// based on the C++ code used in Jikes to determine
// the most specific method. 

// Based on the Jikes
// isSubType( Object, String ) -> true
// isSubType( String, Object) -> false. 
// isSubType (String, Long) -> false. 

private static boolean isSubType(Class superClass, Class subClass ) {
// null is a subType of every type, each type is its own subtype. 
    if( superClass == subClass ) { return true; }
    if( subClass == null ) { return true; }
    if( superClass == null ) { return false; }
// Check interfaces and classes separately. 
    if( superClass.isInterface() ) {
        Class [] theInterfaces = subClass.getInterfaces();
        for( int i = 0; i < theInterfaces.length; i++ ) {
            if( isSubType( superClass, theInterfaces[i] ) ) {
                return true; 
                }
            } 
        return false; 
        }
    else { // superClass is not an interface
        if( subClass.isInterface() ) { return false; }
        for( Class c = subClass.getSuperclass(); c != null ; c = c.getSuperclass() ){
            if( c == superClass ) { return true; }
            }
        return false;
        }
    }
    
     
private static boolean MoreSpecificByObject( Object [] source, Class [] target ) {
    
    for (int k = target.length - 1; k >= 0; k--)
    {
        if( !( source[k] == null || ( target[k].isInstance(source[k])) ) ) {
            return false; 
            }
    }
    return true;
   }  
    
    
private static boolean MoreSpecific( Class [] source, Class [] target ) {
    // Sun bug 4761586: the declaration type is no longer considered when
    // looking for the most specific method.
 
    for (int k = target.length - 1; k >= 0; k--)
    {
        if (!isSubType(target[k], source[k]) ) { return false; } 
    }
    return true;
   }
   
private static boolean NoMethodMoreSpecific ( Vector maximally_specific_method, Class [] method ) 
 {
    for (int i = 0; i < maximally_specific_method.size(); i++)
    {
        if ( MoreSpecific(((Method) maximally_specific_method.get(i)).getParameterTypes(), method))
            return false;
    }
    return true;
}

private static boolean MethodMoreSpecific(Class [] method,
                                   Vector maximally_specific_method)
{
    for (int i = 0; i < maximally_specific_method.size(); i++)
    {
        if (! MoreSpecific(method, ((Method) maximally_specific_method.get(i)).getParameterTypes() ))
            return false;
    }
    return true;
}
/**
Returns a vector of methods of class c and name methodName that are considered as 
"most specific" according to 
the Java specification. If the vector consists of more than one method all methods in the vector are
equally specific (this means that a call to this method is ambigous). Defined 
on package level. 
*/

static Vector getMostSpecificMethod( Class c, String methodName, Object [] parameterList ) {
    Method [] allMethods = c.getMethods();
    Vector methodSet = new Vector();
    for( int i = 0; i < allMethods.length; i++ ) {
        Method m = allMethods[i];
        
        if( m.getName().equals(methodName) ) {
            Class [] params =  m.getParameterTypes();
            if( params.length == parameterList.length &&
                MoreSpecificByObject( parameterList, params ) ) {
                if (MethodMoreSpecific(params, methodSet)){
                    methodSet.clear();
                    methodSet.add(m);
                }
                else if (NoMethodMoreSpecific(methodSet, params)) {
                    methodSet.add(m); 
                }
            }
        }
    }
    return methodSet;
}



/**
*    An instance of  <code>shadow</code> is added whenever an <code>owner</code>object
*    is instantiated. The order in that the shadows
*    are added to the object is the same order as the shadow classes are
*    added to <code>owner</code>. 
*       
*    @param owner The class that is to be shadowed
*    @param shadow The class that shadows <code>owner</code> 
*     <!-- -->
*/     
     public static void addShadow(Class owner, Class shadow) {
           com.perisic.jshadow.Shadowable.addShadow(owner,shadow);
            }
            
/**
*     Removes the class <code>shadow</code> that has been added to the class
*     <code>owner</code>. Note: <em>Objects</em> that already have been instantiated from 
*     the class remain untouched
*     by this method, i.e. shadows are not removed. 
*
*     @param owner the class that is shadowed
*     @param shadow the shadow to be removed from the class
*     @return true if  <code>owner</code> has been shadowed by <code>shadow</code>
*    
*/            
      public static boolean removeShadow(Class owner, Class shadow) {
            return com.perisic.jshadow.Shadowable.removeShadow(owner,shadow);
            }
/**
*    Gets a vector of all shadows that belong to the class <code>owner</code>. 
*    This may be used to manipulate the class shadows manually by changing the
*    the vector.
*    @param owner the class whose shadows are returned
*    @return all classes added to <code>owner</code>  
*/
            
      public static Vector getClassShadows(Class owner) {
        return com.perisic.jshadow.Shadowable.getClassShadows(owner);
            }
           
     
/**
*     Gets all classes that have been shadowed.
*     @return an array of all classes that are shadowed.  
*/        
      public static Class [] allShadowedClasses() {
            return com.perisic.jshadow.Shadowable.allShadowedClasses();
            }

}
     
 
